/*****************************************************************************

This source code is made available to you under the terms of the Open
Development License.  This Package should be accompanied by a file that
contains the associated licensing terms.  If you are unable to find this file,
please refer to its location on the Internet at www.rocklyte.com/license.html.

******************************************************************************

###CLASS###
Name:      Gamma
Version:   1.0
Status:    Beta
Date:      March 2003
ID:        ID_GAMMA
Category:  Effect
Author:    Rocklyte Systems
Copyright: Rocklyte Systems (c) 2003.  All rights reserved.
Short:     .
Keywords:  gamma, correction, red, green, blue, fade, colour

###DESCRIPTION###

<p>To fade to black over one second:</p>

  <gamma fadeout seconds="1.0"/>

<p>To fade in from black over 1/2 a second:</p>

  <gamma fadein seconds="0.5"/>

###END###

*****************************************************************************/

#include <pandora/system/all.h>
#include <pandora/graphics/screen.h>
#include <pandora/functions/strtoint.c>

static ERROR CMDInit(OBJECTPTR, struct KernelBase *);
static ERROR CMDExpunge(void);

#define VER_GAMMA 1.0

EXPORT struct ModHeader ModHeader = {
   MODULE_HEADER_V1, CMDInit, NULL, NULL, CMDExpunge,
   JMP_DEFAULT, 0, CPU_PC, VER_GAMMA, VER_KERNEL,
   "Rocklyte Systems",
   "Rocklyte Systems (c) 2003.  All rights reserved.",
   "March 2003",
   "Gamma"
};

static struct KernelBase *KernelBase = 0;
static OBJECTPTR GammaClass = 0;
static LONG glBreakMessageID = 0;

#define FPS 50

/*****************************************************************************
** Gamma class definition.
*/

enum {
   GML_FADEIN=1,
   GML_FADEOUT
};

struct Gamma {
   OBJECT_HEADER
   LONG  Flags;
   LONG  Static;
   LONG  Wait;
   FLOAT Speed;
   FLOAT Seconds;

   PRIVATE_FIELDS
   OBJECTID ScreenID;
   struct RGB Colour;
   UBYTE Fade;
   FLOAT Counter;
   FLOAT DefaultRed, DefaultGreen, DefaultBlue;
   UBYTE Active;
};

static struct FieldDef GammaFlags[] = {
   { NULL, NULL }
};

static ERROR SET_FadeOut(struct Gamma *, STRING);
static ERROR SET_FadeIn(struct Gamma *, STRING);
static ERROR SET_Seconds(struct Gamma *, FLOAT *);

static struct FieldArray GammaFields[] = {
   { "Flags",    0, FDF_LONGFLAGS|FDF_RW, (LONG)&GammaFlags, NULL, NULL },
   { "Static",   0, FDF_LONG|FDF_RI,      0, NULL, NULL },
   { "Wait",     0, FDF_LONG|FDF_RW,      0, NULL, NULL },
   { "Speed",    0, FDF_FLOAT|FDF_RW,     0, NULL, NULL },
   { "Seconds",  0, FDF_FLOAT|FDF_RW,     0, NULL, SET_Seconds },
   /*** Variable fields ***/
   { "FadeOut",  0, FDF_STRING|FDF_W,     0, NULL, SET_FadeOut },
   { "FadeIn",   0, FDF_STRING|FDF_W,     0, NULL, SET_FadeIn },
   END_FIELD
};

static ERROR GAMMA_Activate(struct Gamma *, APTR);
static ERROR GAMMA_ClosingTag(struct Gamma *, APTR);
static ERROR GAMMA_Free(struct Gamma *, APTR);
static ERROR GAMMA_Init(struct Gamma *, APTR);
static ERROR GAMMA_NewObject(struct Gamma *, APTR);
static ERROR GAMMA_Timer(struct Gamma *, struct acTimer *);

static struct ActionArray GammaActions[] = {
   { AC_Activate,   GAMMA_Activate },
   { AC_ClosingTag, GAMMA_ClosingTag },
   { AC_Free,       GAMMA_Free },
   { AC_Init,       GAMMA_Init },
   { AC_NewObject,  GAMMA_NewObject },
   { AC_Timer,      GAMMA_Timer },
   { NULL, NULL }
};

/****************************************************************************/

static ERROR ReadColour(STRING, LONG *, LONG *, LONG *);

/*****************************************************************************
** Command: Init()
*/

static ERROR CMDInit(OBJECTPTR argModule, struct KernelBase *argKernelBase)
{
   KernelBase = argKernelBase;

   glBreakMessageID = AllocateID(IDTYPE_MESSAGE);

   return(CreateObject(ID_CLASS, NULL, (OBJECTPTR *)&GammaClass, NULL,
      FID_BaseClassID|TLONG, ID_GAMMA,
      FID_Version|TFLOAT,    VER_GAMMA,
      FID_Name|TSTRING,      "Gamma",
      FID_Category|TLONG,    CCF_EFFECT,
      FID_Actions|TPTR,      GammaActions,
      FID_Fields|TPTR,       GammaFields,
      FID_Size|TLONG,        sizeof(struct Gamma),
      TAGEND));
}

/*****************************************************************************
** Command: Expunge()
*/

static ERROR CMDExpunge(void)
{
   if (GammaClass) { Action(AC_Free, GammaClass, NULL); GammaClass = NULL; }
   return(ERR_Okay);
}

/*****************************************************************************
** Gamma: Activate
*/

static ERROR GAMMA_Activate(struct Gamma *Self, APTR Void)
{
   struct mtSetGammaLinear gamma;
   struct Message message;
   struct NextMsg nextmsg;
   OBJECTPTR screen;

   if (Self->Fade IS GML_FADEIN) DPrintF("~Activate()","[Gamma:%d] Fade-In", Self->Head.UniqueID);
   else if (Self->Fade IS GML_FADEOUT) DPrintF("~Activate()","[Gamma:%d] Fade-Out", Self->Head.UniqueID);
   else DPrintF("~Activate()","[Gamma:%d]", Self->Head.UniqueID);

   /* It is very important that this routine is not called recursively
   ** (which can be possible due to the messaging loop a little further
   ** down).
   */

   if (Self->Active IS TRUE) {
      DPrintF("Activate:","[Gamma] Gamma process is already active.");
      StepBack();
      return(ERR_Okay);
   }

   Self->Active = TRUE;

   if (AccessObject(Self->ScreenID, 3000, &screen) IS ERR_Okay) {

      /*** Record the user's gamma settings ***/

      if (GetFields(screen, FID_GammaRed|TFLOAT,   &Self->DefaultRed,
                            FID_GammaGreen|TFLOAT, &Self->DefaultGreen,
                            FID_GammaBlue|TFLOAT,  &Self->DefaultBlue,
                            TAGEND) != ERR_Okay) {
         Self->DefaultRed   = 1.0;
         Self->DefaultGreen = 1.0;
         Self->DefaultBlue  = 1.0;
      }

      ReleaseObject(screen);

      /*** Set the starting gamma values ***/

      if (Self->Fade IS GML_FADEIN) {
         gamma.Red   = 0;
         gamma.Green = 0;
         gamma.Blue  = 0;
         gamma.Flags = NULL;
         ActionMsg(MT_SetGammaLinear, Self->ScreenID, &gamma);
      }
   }

   /*** Initialise the Timer ***/

   SubscribeTimer(Self->Head.UniqueID, 1000 / FPS);

   if (Self->Wait) {
      LARGE oldhandler;

      /* Turn off any message handler that has been setup by the main
      ** program because it would otherwise interfere with our loop.
      */

      oldhandler = SetResource(RES_MESSAGEHANDLER, NULL);

      while (ProcessMessages(&nextmsg, MF_WAIT) IS ERR_Okay) {
         if (nextmsg.Count > 0) {
            if (GetMessage(NULL, NULL, NULL, &message, sizeof(struct Message)) IS ERR_Okay) {
               if ((message.Type IS MSGID_QUIT) OR (message.Type IS glBreakMessageID)) break;
               else DPrintF("@Activate:","[Gamma] Received unknown message type #%d.", message.Type);
            }
         }
      }

      /*** Restore the previous message handler routine ***/

      SetResource(RES_MESSAGEHANDLER, oldhandler);
   }

   Self->Active = FALSE;

   StepBack();
   return(ERR_Okay);
}

/*****************************************************************************
** Gamma: ClosingTag
*/

static ERROR GAMMA_ClosingTag(struct Gamma *Self, APTR Void)
{
   if (Self->Static IS FALSE) {
      Action(AC_Activate, Self, NULL);
   }
   return(ERR_Okay);
}

/*****************************************************************************
** Gamma: Free
*/

static ERROR GAMMA_Free(struct Gamma *Self, APTR Void)
{
   struct mtSetGammaLinear gamma;

   /* If we are being freed before completing our fade process, put the
   ** gamma values back to normal.
   */

   if ((Self->Counter > 0.0) AND (Self->Counter < 1.0)) {
      if (Self->Fade IS GML_FADEIN) {
         gamma.Red   = Self->DefaultRed;
         gamma.Green = Self->DefaultGreen;
         gamma.Blue  = Self->DefaultBlue;
         gamma.Flags = NULL;
         ActionMsg(MT_SetGamma, Self->ScreenID, &gamma);
      }
      else if (Self->Fade IS GML_FADEOUT) {
         gamma.Red   = 0;
         gamma.Green = 0;
         gamma.Blue  = 0;
         gamma.Flags = NULL;
         ActionMsg(MT_SetGammaLinear, Self->ScreenID, &gamma);
      }
   }

   UnsubscribeTimer(Self->Head.UniqueID);
   return(ERR_Okay);
}

/*****************************************************************************
** Gamma: Init
*/

static ERROR GAMMA_Init(struct Gamma *Self, APTR Void)
{
   if (FastFindObject("SystemScreen", ID_SCREEN, &Self->ScreenID, 1, NULL) IS ERR_Okay) {
      return(ERR_Okay);
   }
   else return(ObjectError(Self, ERH_Init, ERR_Search));
}

/*****************************************************************************
** Gamma: NewObject
*/

static ERROR GAMMA_NewObject(struct Gamma *Self, APTR Void)
{
   Self->Speed        = 0.01;
   Self->DefaultRed   = 1.0;
   Self->DefaultGreen = 1.0;
   Self->DefaultBlue  = 1.0;
   return(ERR_Okay);
}

/*****************************************************************************
** Action: Timer
*/

static ERROR GAMMA_Timer(struct Gamma *Self, struct acTimer *Args)
{
   struct mtSetGammaLinear gamma;

   if (!Args) return(ObjectError(Self, ERH_Timer, ERR_Args));

   Self->Counter += Self->Speed;
   if (Self->Counter > 1.0) {
      Self->Counter = 1.0;
      if (Self->Fade IS GML_FADEIN) {
         gamma.Red   = Self->DefaultRed;
         gamma.Green = Self->DefaultGreen;
         gamma.Blue  = Self->DefaultBlue;
         gamma.Flags = NULL;
         ActionMsg(MT_SetGamma, Self->ScreenID, &gamma);
      }
      else if (Self->Fade IS GML_FADEOUT) {
         gamma.Red   = 0;
         gamma.Green = 0;
         gamma.Blue  = 0;
         gamma.Flags = NULL;
         ActionMsg(MT_SetGammaLinear, Self->ScreenID, &gamma);
      }
   }
   else if (Self->Fade IS GML_FADEIN) {
      gamma.Red   = Self->DefaultRed   * Self->Counter;
      gamma.Green = Self->DefaultGreen * Self->Counter;
      gamma.Blue  = Self->DefaultBlue  * Self->Counter;
      gamma.Flags = NULL;
      ActionMsg(MT_SetGammaLinear, Self->ScreenID, &gamma);
   }
   else if (Self->Fade IS GML_FADEOUT) {
      gamma.Red   = Self->DefaultRed   * (1.0 - Self->Counter);
      gamma.Green = Self->DefaultGreen * (1.0 - Self->Counter);
      gamma.Blue  = Self->DefaultBlue  * (1.0 - Self->Counter);
      gamma.Flags = NULL;
      ActionMsg(MT_SetGammaLinear, Self->ScreenID, &gamma);
   }

   if (Self->Counter >= 1.0) {
      UnsubscribeTimer(Self->Head.UniqueID);

      if (Self->Wait) SendMessage(NULL, glBreakMessageID, NULL, NULL, NULL);
      else Action(AC_Free, Self, NULL);
      return(ERR_Okay);
   }

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   FadeIn
Short:  .
Type:   STRING
Status: Set

###DESCRIPTION###

###END###

*****************************************************************************/

static ERROR SET_FadeIn(struct Gamma *Self, STRING Value)
{
   ReadColour(Value, &Self->Colour.Red, &Self->Colour.Green, &Self->Colour.Blue);
   Self->Fade = GML_FADEIN;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   FadeOut
Short:  .
Type:   STRING
Status: Set

###DESCRIPTION###

###END###

*****************************************************************************/

static ERROR SET_FadeOut(struct Gamma *Self, STRING Value)
{
   ReadColour(Value, &Self->Colour.Red, &Self->Colour.Green, &Self->Colour.Blue);
   Self->Fade = GML_FADEOUT;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Seconds
Short:  .
Type:   FLOAT
Status: Read/Set

###DESCRIPTION###

###END###

*****************************************************************************/

static ERROR SET_Seconds(struct Gamma *Self, FLOAT *Value)
{
   Self->Speed = (1.0 / FPS) / *Value;
   return(ERR_Okay);
}

/*****************************************************************************
** Internal: ReadByte()
** Short:    Reads any double digit hex value and converts it into a number.
*/

static WORD ReadByte(STRING Str)
{
   WORD Value;
   if (!*Str) return(0);
   if ((*Str >= '0') AND (*Str <= '9')) Value = (*Str - '0') * 16;
   else if ((*Str >= 'A') AND (*Str <= 'F')) Value = ((*Str - 'A')+10) * 16;
   else if ((*Str >= 'a') AND (*Str <= 'f')) Value = ((*Str - 'a')+10) * 16;
   else Value = 0;
   Str++;
   if ((*Str >= '0') AND (*Str <= '9')) Value += (*Str - '0');
   else if ((*Str >= 'A') AND (*Str <= 'F')) Value += (*Str - 'A') + 10;
   else if ((*Str >= 'a') AND (*Str <= 'f')) Value += (*Str - 'a') + 10;
   return(Value);
}

static ERROR ReadColour(STRING Colour, LONG *Red, LONG *Green, LONG *Blue)
{
   if (*Colour IS '#') {
      Colour++;
      *Red   = ReadByte(Colour);
      if (*Colour) Colour++; if (*Colour) Colour++;
      *Green = ReadByte(Colour);
      if (*Colour) Colour++; if (*Colour) Colour++;
      *Blue  = ReadByte(Colour);
   }
   else {
      *Red = StrToInt(Colour);
      while ((*Colour) AND (*Colour != ',')) Colour++; if (*Colour) Colour++;
      *Green = StrToInt(Colour);
      while ((*Colour) AND (*Colour != ',')) Colour++; if (*Colour) Colour++;
      *Blue = StrToInt(Colour);
   }
   return(ERR_Okay);
}
